#include <Arduino.h>
#include <stdint.h>
#include "pins_arduino.h"
#include <avr/io.h>
#include "sd_io.h"

uint8_t SD_L0_CSPin;

/**
 * Setup pins and SPI for operation with sd-card.
 *
 * Pin settings:
 * - Set all pins for proper operation
 * - power up card if needed
 *
 * SPI configuration: 
 * - enable uC for SPI master
 * - typical no interrupts are used for SPI
 * - data order: MSB is transmitted first
 * - clock polarity: CLK is low when idle
 * - clock phase: 1-0 > Sample, 0-1 > Setup
 * - clock frequency: less than 400kHz 
 *   (will be switched to higher value after initialization)
 *
 */
void SD_L0_Init(void)
{
	/* Power up card */
	// not for arduino
	
	/* Setup ports */
   pinMode(SD_L0_CSPin, OUTPUT);
   SD_L0_SetCSHigh();
   
   // pinMode(SD_L0_SPI_MISO_PIN, INPUT);
   // pinMode(SD_L0_SPI_MOSI_PIN, OUTPUT);
   // pinMode(SD_L0_SPI_SCK_PIN, OUTPUT);
   // pinMode(SD_L0_SPI_SS_FORCE_OUTPUT, OUTPUT);
   
   pinMode(MISO, INPUT);
   pinMode(SCK, OUTPUT);
   pinMode(MOSI, OUTPUT);
   pinMode(SS, OUTPUT);
  
   digitalWrite(SCK, LOW);
   digitalWrite(MOSI, LOW);
   digitalWrite(SS, HIGH);
   
    /* Powering up takes at least 500us for capacitors to charge */
    // not for arduino

    /* initialize SPI with lowest frequency; max. 400kHz during identification mode of card */
    SPCR = (0 << SPIE) | /* SPI Interrupt Enable */
           (1 << SPE)  | /* SPI Enable */
           (0 << DORD) | /* Data Order: MSB first */
           (1 << MSTR) | /* Master mode */
           (0 << CPOL) | /* Clock Polarity: SCK low when idle */
           (0 << CPHA) | /* Clock Phase: sample on rising SCK edge */
           (1 << SPR1) | /* Clock Frequency: f_OSC / 128 */
           (1 << SPR0);
    SPSR &= ~(1 << SPI2X); /* No doubled clock frequency */

}

/**
 * Setup pins and SPI for minimum power consumption.
 * Deactivate SPI and power from card
 */
void SD_L0_DeInit(void)
{
	SPCR &= ~(1 << SPE);
	
	/* Setup ports */
	/* for safe eject */
   pinMode(SD_L0_CSPin, INPUT);
   pinMode(MISO, INPUT);
   pinMode(MOSI, INPUT);
   pinMode(SCK, INPUT);
   digitalWrite(SD_L0_CSPin, HIGH); // Set CS high
   digitalWrite(SCK, HIGH); 
	
	/* Power down card */
	// not for arduino
}

/**
 * Set SPI for full operation speed (up to 25 MHz).
 *
 * Will be called after first part of card 
 * initialization was successful.
 */
void SD_L0_SpiSetHighSpeed(void)
{
    SPCR &= ~((1 << SPR1) | (1 << SPR0)); /* Clock Frequency: f_OSC / 4 */
    SPSR |= (1 << SPI2X);         /* Doubled Clock Frequency: f_OSC / 2 */
}

/**
 * Receives a raw byte from the SPI.
 *
 * \returns The byte which should be read.
 */
uint8_t SD_L0_SpiRecvByte()
{
    /* send dummy data for receiving some */
    SPDR = 0xff;
    while(!(SPSR & (1 << SPIF)));
    SPSR &= ~(1 << SPIF);

    return SPDR;
}

/**
 * Sends a raw byte to the SPI.
 *
 * \param[in] b The byte to sent.
 */
void SD_L0_SpiSendByte(uint8_t b)
{
    SPDR = b;
    /* wait for byte to be shifted out */
    while(!(SPSR & (1 << SPIF)));
    SPSR &= ~(1 << SPIF);
}

/** 
 * SPI read data - only one call so force inline 
 *
 * Receives nbytes in buf. Send 0xff all the time.
 */
//inline __attribute__((always_inline)) 
void SD_L0_SpiRecvBlock(uint8_t *buf, uint16_t nbyte) 
{
  if (nbyte-- == 0) return;
  SPDR = 0xFF;
  for (uint16_t i = 0; i < nbyte; i++) {
    while (!(SPSR & (1 << SPIF)));
    buf[i] = SPDR;
    SPDR = 0xFF;
  }
  while (!(SPSR & (1 << SPIF)));
  buf[nbyte] = SPDR;
}

/** 
 * SPI send block - only one call so force inline 
 *
 * Sends the token and 512 data bytes from buffer afterwards.
 */
//inline __attribute__((always_inline)) 
void SD_L0_SpiSendBlock(uint8_t token, const uint8_t *buf) 
{
  SPDR = token;
  for (uint16_t i = 0; i < 512; i += 2) {
    while (!(SPSR & (1 << SPIF)));
    SPDR = buf[i];
    while (!(SPSR & (1 << SPIF)));
    SPDR = buf[i + 1];
  }
  while (!(SPSR & (1 << SPIF)));
}





//################################################################################################################################################







//#include "sd_l0.h"
//#include "sd_l1.h"

// SD card commands
/** GO_IDLE_STATE - init card in spi mode if CS low */
#define SD_CMD0 0x00
/** SEND_IF_COND - verify SD Memory Card interface operating condition.*/
#define SD_CMD8 0x08
/** SEND_CSD - read the Card Specific Data (CSD register), response R1 */
#define SD_CMD9 0x09
/** SEND_CID - read the card identification information (CID register), response R1 */
#define SD_CMD10 0x0A
/** STOP_TRANSMISSION - end multiple block read sequence, response R1b */
#define SD_CMD12 0x0C
/** SEND_STATUS - read the card status register, response R2 */
#define SD_CMD13 0x0D
/** SET_BLOCKLEN arg0[31:0]: block length, response R1 */
#define SD_CMD16 0x10
/** READ_SINGLE_BLOCK - read a single data block from the card, response R1 */
#define SD_CMD17 0x11
/** READ_MULTIPLE_BLOCK - read a multiple data blocks from the card, response R1 */
#define SD_CMD18 0x12
/** WRITE_BLOCK - write a single data block to the card, response R1 */
#define SD_CMD24 0x18
/** WRITE_MULTIPLE_BLOCK - write blocks of data until a STOP_TRANSMISSION, response R1 */
#define SD_CMD25 0x19
/** ERASE_WR_BLK_START - sets the address of the first block to be erased */
//#define SD_CMD32 0x20
/** ERASE_WR_BLK_END - sets the address of the last block of the continuous
    range to be erased*/
//#define SD_CMD33 0x21
/** ERASE - erase all previously selected blocks */
//#define SD_CMD38 0x26
/** APP_CMD - escape for application specific command */
#define SD_CMD55 0x37
/** READ_OCR - read the OCR register of a card */
#define SD_CMD58 0x3A
/** CRC_ON_OFF - Turns CRC option on or off, response R1 */
#define SD_CMD59 0x3B

/** SET_WR_BLK_ERASE_COUNT - Set the number of write blocks to be
     pre-erased before writing, response R1 */
#define SD_ACMD23 0x17
/** SD_SEND_OP_COMD - Sends host capacity support information and
    activates the card's initialization process */
#define SD_ACMD41 0x29

//------------------------------------------------------------------------------
/** status for card in the ready state */
#define SD_R1_READY_STATE 0x00
/** status for card in the idle state */
#define SD_R1_IDLE_STATE 0x01
/** status bit for illegal command */
#define SD_R1_ILLEGAL_COMMAND 0x04
/** start data token for read or write single block*/
#define SD_DATA_START_BLOCK 0xFE
/** stop token for write multiple blocks*/
//#define SD_STOP_TRAN_TOKEN 0xFD
/** start data token for write multiple blocks*/
//#define SD_WRITE_MULTIPLE_TOKEN 0xFC
/** mask for data response tokens after a write block operation */
//#define SD_DATA_RES_MASK 0x1F
/** write data accepted token */
//#define SD_DATA_RES_ACCEPTED 0x05

// prototypes for internal usage
uint8_t SD_L1_WaitNotBusy(uint16_t timeout);
uint8_t SD_L1_CardCommand(uint8_t cmd, uint32_t arg);
uint8_t SD_L1_CardACommand(uint8_t cmd, uint32_t arg);
uint8_t SD_L1_ReadData(uint8_t *dst, uint16_t count);
uint8_t SD_L1_WriteData(uint8_t token, const uint8_t *src);

uint8_t SD_L1_type;

/** Sends also one dummy byte to ensure MISO goes high impedance */
void SD_L1_SetCSHigh() {
  SD_L0_SetCSHigh();
  SD_L0_SpiSendByte(0xff);
}

#define SD_L1_SetCSLow() SD_L0_SetCSLow()

/** wait for card to go not busy */
uint8_t SD_L1_WaitNotBusy(uint16_t timeout) 
{
  uint16_t t0 = SD_L0_GetTimestamp();
  while (SD_L0_SpiRecvByte() != 0xFF) {
    if ((SD_L0_GetTimestamp() - t0) >= timeout) return 0;
  }
  return 1;
}

/**
 * Send a command to the memory card which responses with a R1 response (and possibly others).
 *
 * \param[in] command The command to send.
 * \param[in] arg The argument for command.
 * \returns The command answer.
 */
uint8_t SD_L1_CardCommand(uint8_t cmd, uint32_t arg) 
{
  uint8_t response;
  uint8_t crc = 0xFF;
      
  // select card
  SD_L1_SetCSLow();

  // wait up to timeout if busy
  SD_L1_WaitNotBusy(SD_COMMAND_TIMEOUT);

  // send command
  SD_L0_SpiSendByte(cmd | 0x40);

  // send argument
  SD_L0_SpiSendByte((arg >> 24) & 0xff);
  SD_L0_SpiSendByte((arg >> 16) & 0xff);
  SD_L0_SpiSendByte((arg >> 8) & 0xff);
  SD_L0_SpiSendByte((arg >> 0) & 0xff);
  
  // send CRC, only required for commands 0 and 8
  if (cmd == SD_CMD0) crc = 0x95;  // correct crc for CMD0 with arg 0
  if (cmd == SD_CMD8) crc = 0x87;  // correct crc for CMD8 with arg 0X1AA
  SD_L0_SpiSendByte(crc);

  // skip stuff byte for stop read
  if (cmd == SD_CMD12) SD_L0_SpiRecvByte();

  // wait for response
  for(uint8_t i = 0; i < 100; ++i)
  {
      response = SD_L0_SpiRecvByte();
      if(response != 0xff)
          break;
  }
  return response;
}

/**
 * Send an application specific command which responses with a R1 response (and possibly others).
 *
 * \param[in] command The command to send.
 * \param[in] arg The argument for command.
 * \returns The command answer.
 */
uint8_t SD_L1_CardACommand(uint8_t cmd, uint32_t arg)
{
    SD_L1_CardCommand(SD_CMD55, 0);
    return SD_L1_CardCommand(cmd, arg);
}

/**
 * Read one block of data from card.
 *
 * \return Error code, 0 if ok.
 */
uint8_t SD_L1_ReadData(uint8_t *dst, uint16_t count) 
{
  // wait for start block token
  uint8_t status;
  uint16_t t0 = SD_L0_GetTimestamp();
  
  while ((status = SD_L0_SpiRecvByte()) == 0xFF) {
    if ((SD_L0_GetTimestamp() - t0) > SD_READ_TIMEOUT) {
      SD_L1_SetCSHigh();
      return(SD_CARD_ERROR_READ_TIMEOUT);
    }
  }
  
  if (status != SD_DATA_START_BLOCK) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_READ);
  }
  
  // transfer data
  SD_L0_SpiRecvBlock(dst, count);

  // discard CRC
  SD_L0_SpiRecvByte();
  SD_L0_SpiRecvByte();
  SD_L1_SetCSHigh();

  return 0;
}

/**
 * Write one block of data to card.
 *
 * \return Error code, 0 if ok.
 */
uint8_t SD_L1_WriteData(uint8_t token, const uint8_t *src) 
{
  uint8_t status;
  
  SD_L0_SpiSendBlock(token, src);

  SD_L0_SpiSendByte(0xff);  // dummy crc
  SD_L0_SpiSendByte(0xff);  // dummy crc

  status = SD_L0_SpiRecvByte();
  if ((status & 0x1f) != 0x05) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_WRITE);
  }
  return 0;
}

// --------------------------------------------------------------------------------------
// ------------------------------------- Public functions -------------------------------
// --------------------------------------------------------------------------------------

/**
 * Initialize the SD memory card.
 *
 * Power up the card, set SPI mode.
 * Detects the card version (V1, V2, SDHC), sets
 * sector length to 512.
 *
 * \return Zero if successfull, errorcode otherwise
 */
uint8_t SD_L1_Init()
{
  uint8_t status;
  // 16-bit init start time allows over a minute
  uint16_t t0 = SD_L0_GetTimestamp();
  uint32_t arg;
  
  SD_L1_type = 0;

  // set pin modes
  SD_L0_Init();
  
  // must supply min of 74 clock cycles with CS high.
  SD_L1_SetCSHigh();
  for (uint8_t i = 0; i < 10; i++) SD_L0_SpiSendByte(0xFF);

  // command to go idle in SPI mode
  while ((SD_L1_CardCommand(SD_CMD0, 0)) != SD_R1_IDLE_STATE) {
    if ((SD_L0_GetTimestamp() - t0) > SD_INIT_TIMEOUT) {
      SD_L1_SetCSHigh();
      return(SD_CARD_ERROR_CMD0);
    }
  }
  
  // check SD version ( 2.7V - 3.6V + test pattern )
  if ((SD_L1_CardCommand(SD_CMD8, 0x1AA) & SD_R1_ILLEGAL_COMMAND)) {
    SD_L1_type = SD_CARD_TYPE_SD1;
    // Not done here: Test if SD or MMC card here using CMD55 + CMD1
  } else {
    // only need last byte of r7 response
    SD_L0_SpiRecvByte();
    SD_L0_SpiRecvByte();
    status = SD_L0_SpiRecvByte();
    if ((status & 0x01) == 0) {
      // card operation voltage range doesn't match
      SD_L1_SetCSHigh();
      return(SD_CARD_ERROR_VOLTMATCH);       
    }
    if (SD_L0_SpiRecvByte() != 0xAA) {
      SD_L1_SetCSHigh();
      return(SD_CARD_ERROR_CMD8);
    }
    SD_L1_type = SD_CARD_TYPE_SD2;
  }
  
  // Turn CRC option off
  SD_L1_CardCommand(SD_CMD59, 0);
  
  // initialize card and send host supports SDHC if SD2
  arg = (SD_L1_type == SD_CARD_TYPE_SD2) ? 0X40000000 : 0;
  while ((SD_L1_CardACommand(SD_ACMD41, arg)) != SD_R1_READY_STATE) {
    // check for timeout
    if ((SD_L0_GetTimestamp() - t0) > SD_INIT_TIMEOUT) {
      SD_L1_SetCSHigh();
      return(SD_CARD_ERROR_ACMD41);
    }
  }
  
  // if SD2 read OCR register to check for SDHC card
  if (SD_L1_type == SD_CARD_TYPE_SD2) {
    if (SD_L1_CardCommand(SD_CMD58, 0)) {
      SD_L1_SetCSHigh();
      return(SD_CARD_ERROR_CMD58);
    }
    // other implementation test only against 0x40 for SDHC detection...
    if ((SD_L0_SpiRecvByte() & 0xC0) == 0xC0) SD_L1_type = SD_CARD_TYPE_SDHC;
    // discard rest of ocr - contains allowed voltage range
    SD_L0_SpiRecvByte();
    SD_L0_SpiRecvByte();
    SD_L0_SpiRecvByte();
  }

  // set block size to 512 bytes
  if(SD_L1_CardCommand(SD_CMD16, 512))
  {
      SD_L1_SetCSHigh();
      return(SD_CARD_ERROR_CMD16);
  }
  
  SD_L1_SetCSHigh();
  SD_L0_SpiSetHighSpeed();
  return 0;
}

/**
 * Shutdown SD card safely 
 */
void SD_L1_DeInit()
{
  SD_L1_type = 0;
  
  // must supply min of 74 clock cycles with CS high.
  SD_L1_SetCSHigh();
  for (uint8_t i = 0; i < 10; i++) SD_L0_SpiSendByte(0xFF);
  
  SD_L0_DeInit();
}

/** Return the card type: SD V1, SD V2 or SDHC
 * \return 0 - no card, 1 - SD V1, 2 - SD V2, or 3 - SDHC.
 */
uint8_t SD_L1_GetCardType() 
{
    return(SD_L1_type);
}

/**
 * Read a 512 byte block from an SD card.
 *
 * \param[in] blockNumber Logical block to be read.
 * \param[out] dst Pointer to the location that will receive the data.
 *
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_ReadBlock(uint32_t blockNumber, uint8_t *dst) 
{
  // use address if not SDHC card
  if (SD_L1_type != SD_CARD_TYPE_SDHC) blockNumber <<= 9;
  if (SD_L1_CardCommand(SD_CMD17, blockNumber)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_CMD17);
  }
  return SD_L1_ReadData(dst, 512);
}

#if SD_ENABLE_MULTIBLOCK_ACCESS
/** 
 * Start a read multiple blocks sequence.
 *
 * \param[in] blockNumber Address of first block in sequence.
 *
 * \note This function is used with ReadMB() and StopReadMB() for optimized
 * multiple block reads.  SPI chipSelect must be low for the entire sequence.
 *
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_ReadMBStart(uint32_t blockNumber) 
{
  if (SD_L1_type != SD_CARD_TYPE_SDHC) blockNumber <<= 9;
  if (SD_L1_CardCommand(SD_CMD18, blockNumber)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_CMD18);
  }
  SD_L1_SetCSHigh();
  return 0;
}

/** 
 * Read one data block in a multiple block read sequence
 *
 * \param[in] dst Pointer to the location that will receive the data.
 *
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_ReadMB(uint8_t *dst) 
{
  SD_L1_SetCSLow();
  return SD_L1_ReadData(dst, 512);
}

/** 
 * End a read multiple blocks sequence.
 *
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_ReadMBStop() 
{
  SD_L1_SetCSLow();
  if (SD_L1_CardCommand(SD_CMD12, 0)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_CMD12);
  }
  SD_L1_SetCSHigh();
  return 0;  
}
#endif /* SD_ENABLE_MULTIBLOCK_ACCESS */

#if SD_ENABLE_L1_INFORMATIVE
/**
 * Read the 16 byte CSD register of SD card.
 *
 * Remark: It seems that the only usefull information stored in CSD
 *         is the actual card size only as other fields are 
 *         typically filled with too general values.
 *
 * \param[out] dst Pointer to the location that will receive the data.
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_ReadCSD(uint8_t *dst)
{
  if (SD_L1_CardCommand(SD_CMD9, 0)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_CMD9);
  }
  return SD_L1_ReadData(dst, 16);
}

/**
 * Read the 16 byte CID register of SD card.
 *
 * \param[out] dst Pointer to the location that will receive the data.
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_ReadCID(uint8_t *dst)
{
  if (SD_L1_CardCommand(SD_CMD10, 0)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_CMD10);
  }
  return SD_L1_ReadData(dst, 16);
} 

/*        
 SD_L1_GetCardType: 2, ReadCSD: 00 7f ff 32 5b 59 83 b9  f6 db df ff 12 40 00 8f  
 SD_L1_GetCardType: 3, ReadCSD: 40 0e 00 32 5b 59 00 00  3b 33 7f 80 0a 40 00 a5  
*/
/**
 * Determine the size of an SD flash memory card.
 *
 * \return The number of 512 byte data blocks in the card
 *         or zero if an error occurs.
 */
uint32_t SD_L1_GetCardSize() 
{
  uint8_t csddata[16];
  if (SD_L1_ReadCSD(&csddata[0])) return 0;
  if ((csddata[0] & 0xc0) == 0) {
    // version 1.00 cards
    uint8_t read_bl_len = ((csddata[5] & 0x0f));
    uint16_t c_size = ((csddata[6] & 0x03) << 10)
                      | (csddata[7] << 2) 
                      | ((csddata[8] & 0xc0)>>6);
    uint8_t c_size_mult = ((csddata[9] & 0x03) << 1)
                          | ((csddata[10] & 0x80)>>7);
    return (uint32_t)(c_size + 1) << (c_size_mult + read_bl_len - 7);
  } else if ((csddata[0] & 0xc0) == 0x40) {
    // version 2.00 cards
    uint32_t c_size = (((uint32_t)csddata[7] & 0xfc) << 14)
                      | (csddata[8] << 8) 
                      | csddata[9];
    return (c_size + 1) << 10;
  } else {
    // error(SD_CARD_ERROR_BAD_CSD);
    return 0;
  }
}
#endif

#if SD_ENABLE_WRITE_ACCESS
/**
 * Writes a 512 byte block to an SD card.
 *
 * \param[in] blockNumber Logical block to be written.
 * \param[in] src Pointer to the location of the data to be written.
 *
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_WriteBlock(uint32_t blockNumber, const uint8_t *src) 
{
  uint8_t response;

  // use address if not SDHC card
  if (SD_L1_type != SD_CARD_TYPE_SDHC) blockNumber <<= 9;
  if (SD_L1_CardCommand(SD_CMD24, blockNumber)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_CMD24);
  }
  
  // Write single block start block token (0xfe), data block and dummy crc to card
  response = SD_L1_WriteData(0xfe, src);
  if (response) return(response);

  // wait for flash programming to complete
  if (!SD_L1_WaitNotBusy(SD_WRITE_TIMEOUT)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_WRITE_TIMEOUT);
  }
  
  // response is r2 so get and check two bytes for nonzero
  if (SD_L1_CardCommand(SD_CMD13, 0) || SD_L0_SpiRecvByte()) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_WRITE_PROGR);
  }
  SD_L1_SetCSHigh();
  return 0;
}

#if SD_ENABLE_MULTIBLOCK_ACCESS
/** 
 * Start a write multiple blocks sequence.
 *
 * \param[in] blockNumber Address of first block in sequence.
 * \param[in] eraseCount The number of blocks to be pre-erased.
 *            Set to 0 if not used.
 *
 * \note This function is used with WriteMB() and StopWriteMB() for optimized
 * multiple block writes.  SPI chipSelect must be low for the entire sequence.
 *
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_WriteMBStart(uint32_t blockNumber, uint32_t eraseCount) 
{
  // send pre-erase count
  if(eraseCount) {
      if (SD_L1_CardACommand(SD_ACMD23, eraseCount)) {
        SD_L1_SetCSHigh();
        return(SD_CARD_ERROR_ACMD23);
      }
  }
  // use address if not SDHC card
  if (SD_L1_type != SD_CARD_TYPE_SDHC) blockNumber <<= 9;
  if (SD_L1_CardCommand(SD_CMD25, blockNumber)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_CMD25);
  }
  SD_L1_SetCSHigh();
  return 0;
}

/** 
 * Write one data block in a multiple block write sequence
 *
 * \param[in] src Pointer to the location of the data to be written.
 *
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_WriteMB(const uint8_t* src) 
{
  SD_L1_SetCSLow();
  // wait for previous write to finish
  if (!SD_L1_WaitNotBusy(SD_WRITE_TIMEOUT)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_WRITE_MB);
  }
  // 0xfc = WRITE_MULTIPLE_TOKEN
  if (SD_L1_WriteData(0xfc, src)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_WRITE_MB);
  }
  SD_L1_SetCSHigh();
  return 0;
}

/** 
 * End a write multiple blocks sequence.
 *
 * \return 0 is returned for success, error code otherwise
 */
uint8_t SD_L1_WriteMBStop() 
{
  SD_L1_SetCSLow();
  if (!SD_L1_WaitNotBusy(SD_WRITE_TIMEOUT)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_STOP_TRAN);
  }
  SD_L0_SpiSendByte(0xfd); // 0xfd = STOP_TRAN_TOKEN
  if (!SD_L1_WaitNotBusy(SD_WRITE_TIMEOUT)) {
    SD_L1_SetCSHigh();
    return(SD_CARD_ERROR_STOP_TRAN);
  }
  SD_L1_SetCSHigh();
  return 0; 
}
#endif /* SD_ENABLE_MULTIBLOCK_ACCESS */
#endif /* SD_ENABLE_WRITE_ACCESS */

//------------------------------------------------------------------------------




//################################################################################################################################################






//#include "sd_l0.h"
//#include "sd_l1.h"
//#include "sd_l2.h"

// #include "dbg_uart.h"
#include <stdio.h>

uint8_t  *SD_L2_workBuf;
SD_L2_FAT_t SD_L2_FAT;

// This both will share same buffer -> NextCluster not public availible!
uint8_t  *SD_L2_FatCache;
uint32_t SD_L2_LastFatSect;

uint8_t     SD_L2_NextCluster(uint32_t *cluster);

/**
 * Returns the first sector of a given cluster
 */
uint32_t SD_L2_Cluster2Sector(uint32_t cluster)
{
    return((cluster << SD_L2_FAT.ClusterSizeShift) + SD_L2_FAT.DataStart);
}

/** 
 * Returns next cluster of a given cluster
 *
 * Param Pointer to cluster number (will be swapped against next cluster)
 *
 * return Zero if successful, error code otherwise
 */
uint8_t SD_L2_NextCluster(uint32_t *cluster)
{
    uint32_t sector;
    uint16_t offset;
    uint8_t  retval;
    
    // calculate sector where cluster information can be found
    if(SD_L2_FAT.PartType == SD_L2_PARTTYPE_FAT16) {
        sector = (*cluster >> 8) + SD_L2_FAT.FatStart;
        offset = (uint16_t)(*cluster & 0xff) << 1;
    } else if(SD_L2_FAT.PartType == SD_L2_PARTTYPE_FAT32) {
        sector = (*cluster >> 7) + SD_L2_FAT.FatStart;
        offset = (uint16_t)(*cluster & 0x7f) << 2;
    } else {
        return(SD_L2_ERROR_FAT_NOT_INIT);
    }
    
    // Read FAT sector if not already buffered
    if(sector != SD_L2_LastFatSect) {
        retval = SD_L1_ReadBlock(sector, SD_L2_FatCache);
        if(retval) return(retval);
        SD_L2_LastFatSect = sector;
    }
    
    // Get FAT entry
    if(SD_L2_FAT.PartType == SD_L2_PARTTYPE_FAT32) {
        *cluster = ((uint32_t)SD_L2_FatCache[offset+3] << 24U) | ((uint32_t)SD_L2_FatCache[offset+2] << 16U);
    } else {
        *cluster = 0;
    }
    *cluster += ((uint32_t)SD_L2_FatCache[offset+1] << 8U) | ((uint32_t)SD_L2_FatCache[offset]);
    
    return(0);
}

/**
 * Test if file is completely not fragmented.
 *
 * return Zero if successful (=not fragmented), error code otherwise
 */
uint8_t SD_L2_IsFileFragmented(SD_L2_File_t *fileinfo)
{
    uint32_t startCluster = fileinfo->FirstCluster;
    uint32_t oldclus = startCluster;
    uint32_t bytecount = 0;
    uint32_t bytepercluster = 512 * SD_L2_FAT.SecPerClus;
    uint8_t  retval = 0;
    SD_L2_FatCache = SD_L2_workBuf;
    SD_L2_LastFatSect = 0xffffffffUL;
    while(retval == 0) {
        retval = SD_L2_NextCluster(&startCluster);
        if(retval) return(retval);
        if(oldclus != (startCluster-1)) {
            retval = SD_L2_ERROR_FRAGMET_FOUND;
        }
        oldclus = startCluster;
        bytecount += bytepercluster;
    }
    if(bytecount < fileinfo->Size) return(retval);
    if(startCluster >= SD_L2_FAT.ClusterEndMarker) {
        return(0);
    } else {
        return(SD_L2_ERROR_FAT_ENTRY);
    }
}

/**
 * DeInitialize the file system (for safe power down mode)
 *
 * Get status (initialized or not) by reading SD_L2_FAT.PartType.
 */
void SD_L2_DeInit()
{
    SD_L2_FAT.PartType = SD_L2_PARTTYPE_UNKNOWN;
    SD_L1_DeInit();
}

/**
 * Initialize the file system .
 *
 * Does the lower level initialization and
 * tries to find the boot sector of the first FAT16
 * or FAT32 partition and parse it.
 * Workbuf must hold at least 512 bytes.
 * Workbuf will be used later also for following functions:
 * - SD_L2_SearchFile
 * - SD_L2_Dir
 *
 * \return Zero if successful, error code otherwise
 */
uint8_t SD_L2_Init(uint8_t *pWorkBuf)
{
    uint8_t  retval;
    uint8_t  PartType;
    uint16_t temp16;
    uint32_t temp32;
    
    SD_L2_FAT.PartType = SD_L2_PARTTYPE_UNKNOWN;
    SD_L2_LastFatSect = 0;
    
    if(pWorkBuf == NULL) return(SD_L2_ERROR_WORKBUF);
    SD_L2_workBuf = pWorkBuf;
    
    // Try init SD-Card
    retval = SD_L1_Init();
    if(retval) return(retval);
    
    // ==== MBR (partition table) access here =====
    
    // Read sector 0 
    retval = SD_L1_ReadBlock(0, SD_L2_workBuf);
    if(retval) return(retval);

    // Test for signature (valid not only for MBR, but FAT Boot Sector as well!)
    if((SD_L2_workBuf[0x1fe] != 0x55) || (SD_L2_workBuf[0x1ff] != 0xaa)) return(SD_L2_ERROR_INVAL_SECT0);
    
    // Store most important MBR values for first partition
    PartType = SD_L2_workBuf[0x1be + 0x04];
    SD_L2_FAT.BootSectorStart =  (uint32_t)SD_L2_workBuf[0x1be + 0x08] 
                             | ((uint32_t)SD_L2_workBuf[0x1be + 0x09] << 8UL)
                             | ((uint32_t)SD_L2_workBuf[0x1be + 0x0a] << 16UL)
                             | ((uint32_t)SD_L2_workBuf[0x1be + 0x0b] << 24UL);
    
    // Check MBR values for plausibility
    if(  ((SD_L2_workBuf[0x1be] & 0x7f) == 0)
      && ((PartType == 0x04) || (PartType == 0x06) || (PartType == 0x0B) 
           || (PartType == 0x0C) || (PartType == 0x0E)) )  
    {
        // MBR seems to contain valid FAT16/FAT32 partition entry
        SD_L2_FAT.PartType = ((PartType == 0x0B) || (PartType == 0x0C)) ? SD_L2_PARTTYPE_FAT32 : SD_L2_PARTTYPE_FAT16;
    }
    else
    {
        // MBR seems to contain not an valid entry, so try for super-floppy now
        SD_L2_FAT.BootSectorStart = 0UL;
        SD_L2_FAT.PartType = SD_L2_PARTTYPE_SUPERFLOPPY;
    }
    
    // ====== FAT access here ======
    
    // Read Boot-Sector and test for signature
    retval = SD_L1_ReadBlock(SD_L2_FAT.BootSectorStart, SD_L2_workBuf);
    if(retval) return(retval);  

    // Test for signature (valid not only for MBR, but FAT Boot Sector as well!)
    if((SD_L2_workBuf[0x1fe] != 0x55) || (SD_L2_workBuf[0x1ff] != 0xaa)) return(SD_L2_ERROR_INVAL_BS);
    
    // Plausibility checks for FAT
    if((SD_L2_workBuf[0x0b] != 0x00) || (SD_L2_workBuf[0x0c] != 0x02) || (SD_L2_workBuf[0x15] != 0xf8)) return(SD_L2_ERROR_INVAL_BS);

    // Read fields that are same for FAT16 and FAT32
    SD_L2_FAT.SecPerClus = SD_L2_workBuf[0x0d];
    SD_L2_FAT.RsvdSecCnt = (uint16_t)SD_L2_workBuf[0x0e] | ((uint16_t)SD_L2_workBuf[0x0f]<<8U);
    if((SD_L2_FAT.SecPerClus == 0) || (SD_L2_FAT.RsvdSecCnt == 0)) return(SD_L2_ERROR_INVAL_BS);
    SD_L2_FAT.NumFATs = SD_L2_workBuf[0x10];
    SD_L2_FAT.RootEntryCount = (uint16_t)SD_L2_workBuf[0x11] | ((uint16_t)SD_L2_workBuf[0x12]<<8U);
    
    temp16 = (uint16_t)SD_L2_workBuf[0x13] | ((uint16_t)SD_L2_workBuf[0x14]<<8U);
    temp32 = (uint32_t)SD_L2_workBuf[0x20] | ((uint32_t)SD_L2_workBuf[0x21]<<8U) | ((uint32_t)SD_L2_workBuf[0x22]<<16U) | ((uint32_t)SD_L2_workBuf[0x23]<<24U);
    SD_L2_FAT.TotalSec  = temp16 ? temp16 : temp32;
    
    temp16 = (uint16_t)SD_L2_workBuf[0x16] | ((uint16_t)SD_L2_workBuf[0x17]<<8U);
    temp32 = (uint32_t)SD_L2_workBuf[0x24] | ((uint32_t)SD_L2_workBuf[0x25]<<8U) | ((uint32_t)SD_L2_workBuf[0x26]<<16U) | ((uint32_t)SD_L2_workBuf[0x27]<<24U);
    SD_L2_FAT.SecPerFAT  = temp16 ? temp16 : temp32;
    
    // Calculate start sectors
    SD_L2_FAT.FatStart = SD_L2_FAT.BootSectorStart + (uint32_t)SD_L2_FAT.RsvdSecCnt;
    SD_L2_FAT.RootDirStart = SD_L2_FAT.FatStart + SD_L2_FAT.NumFATs * (uint32_t)SD_L2_FAT.SecPerFAT;
    
    // Data area starts at cluster #2
    SD_L2_FAT.DataStart = SD_L2_FAT.RootDirStart+ ((32 * (uint32_t)SD_L2_FAT.RootEntryCount + 511)/512) - (2 * SD_L2_FAT.SecPerClus);
    
    // determine shift that is same as multiply by SD_L2_FAT.SecPerClus
    SD_L2_FAT.ClusterSizeShift = 0;
    while (SD_L2_FAT.SecPerClus != (1 << SD_L2_FAT.ClusterSizeShift)) {
      // error if not power of 2
      if (SD_L2_FAT.ClusterSizeShift++ > 7) return(SD_L2_ERROR_INVAL_BS);
    }  
    
     // Calculate number and shifting of clusters
    // total data blocks
    SD_L2_FAT.ClusterCount = SD_L2_FAT.TotalSec - (SD_L2_FAT.DataStart - SD_L2_FAT.BootSectorStart);
    // divide by cluster size to get cluster count
    SD_L2_FAT.ClusterCount >>= SD_L2_FAT.ClusterSizeShift;  
    
    // determine if FAT16 or FAT32 (only by cluster count as done by M$)
    if (SD_L2_FAT.ClusterCount < 4085) {
        // this would be FAT12, which is not supported
        SD_L2_FAT.PartType = SD_L2_PARTTYPE_UNKNOWN;
        return(SD_L2_ERROR_FAT12);
    } else if (SD_L2_FAT.ClusterCount < 65525) {
        SD_L2_FAT.PartType = SD_L2_PARTTYPE_FAT16;
        SD_L2_FAT.ClusterEndMarker = 0xfff8UL;
    } else {
        temp32 = (uint32_t)SD_L2_workBuf[0x2c] | ((uint32_t)SD_L2_workBuf[0x2d]<<8U) | ((uint32_t)SD_L2_workBuf[0x2e]<<16U) | ((uint32_t)SD_L2_workBuf[0x2f]<<24U);
        SD_L2_FAT.RootDirStart = SD_L2_Cluster2Sector(temp32);
        SD_L2_FAT.PartType = SD_L2_PARTTYPE_FAT32;
        SD_L2_FAT.ClusterEndMarker = 0xffffff8UL;
    }
      
    return 0;
}

/**
 * Search a file in the directory.
 * Filename must be 8.3 format, terminated by \0 
 * (can not access ".." now...)
 *
 * Works only over one cluster of directory information. If 
 * SD_L2_ERROR_DIR_EOC is returned call function again with next
 * cluster number. 
 *
 * Set cluster to 0 to access root directory.
 *
 * Deleted files and long name entries are not shown generally.
 *
 * Only files are printed that has their attributes set/unset regarding maskSet/maskUnset.
 * Examples for maskSet, maskUnset:
 *  Ouput everything:           0x00, 0x00
 *  Shortname files only:       0x00, 0x18
 *  Shortname files and dirs:   0x00, 0x08
 *  Shortname dirs:             0x10, 0x08
 *  Volume name:                0x08, 0x10
 *
 * Mask bits: B7 = 0, B6 = 0, B5 = archive, B4 = directory, 
 *            B3 = volume name, B2 = system, B1 = hidden, B0 = read only
 * 
 * If file is found, fileinfo gets filled with file attributes, file size in bytes and first cluster.
 *
 *  return Zero if successfully, error code otherwise
 */
uint8_t SD_L2_SearchFile(uint8_t *filename, const uint32_t cluster, const uint8_t maskSet, const uint8_t maskUnset, SD_L2_File_t *fileinfo)
{
    uint16_t maxsect = SD_L2_FAT.SecPerClus;
    uint32_t startsect = SD_L2_Cluster2Sector(cluster);
    char fnentry[12];
    
    if((SD_L2_FAT.PartType != SD_L2_PARTTYPE_FAT16) && (SD_L2_FAT.PartType != SD_L2_PARTTYPE_FAT32)) return(SD_L2_ERROR_FAT_NOT_INIT);
    if(cluster == 0) {
        // Set root dir sector
        startsect = SD_L2_FAT.RootDirStart;
        if(SD_L2_FAT.PartType == SD_L2_PARTTYPE_FAT16) maxsect = (uint16_t)((32 * (uint32_t)SD_L2_FAT.RootEntryCount + 511)/512);
    }
    
    // convert filename to space-filled uppercase format
    for(uint8_t i = 0; i < 11; i++) fnentry[i] = ' ';
    for(uint8_t i = 0; i < 9; i++) {
        uint8_t c = *filename++;
        if((c < 0x20) || (c == '.')) break;
        if((c>='a') && (c<='z')) c -= 0x20;  // to upper case
        fnentry[i] = c;
    }
    for(uint8_t i = 8; i < 11; i++) {
        uint8_t c = *filename++;
        if(c < 0x20) break;
        if((c>='a') && (c<='z')) c -= 0x20;  // to upper case
        fnentry[i] = c;
    }
    fnentry[11] = 0;
    //Serial.println(fnentry);
    
    // go through sectors
    for(uint16_t i = 0; i<maxsect; i++) {
        uint8_t retval = SD_L1_ReadBlock(startsect + i, SD_L2_workBuf);
        if(retval) return(retval);
        
        for(uint16_t j = 0; j<512; j+=32) {
            uint8_t attrib;
            if(SD_L2_workBuf[j] == 0) return(SD_L2_ERROR_FILE_NOT_FOUND);    // Last entry when first character of filename == 0
            if(SD_L2_workBuf[j] == 0xe5) continue;  // Skip deleted files
            if(SD_L2_workBuf[j] == 0x05) SD_L2_workBuf[j] = 0xE5;
            
            attrib = SD_L2_workBuf[j+0x0b];
            
            // Test masks (skip long file name entries also)
            if(((attrib & maskSet) == maskSet) && ((attrib & maskUnset) == 0) && (attrib != 0x0f)) {
                uint16_t k;
                
                // compare filename
                for(k = 0; k < 11; k++) if(SD_L2_workBuf[j+k] != fnentry[k]) break;
                if(k >= 11) {
                    // found it
                    fileinfo->Attributes = attrib;
                    fileinfo->Size = (uint32_t)SD_L2_workBuf[j+0x1c] | (((uint32_t)SD_L2_workBuf[j+0x1d])<<8) 
                                    | (((uint32_t)SD_L2_workBuf[j+0x1e])<<16) | (((uint32_t)SD_L2_workBuf[j+0x1f])<<24);
                    if(SD_L2_FAT.PartType == SD_L2_PARTTYPE_FAT16) {
                        fileinfo->FirstCluster = (uint32_t)SD_L2_workBuf[j+0x1a] | (((uint32_t)SD_L2_workBuf[j+0x1b])<<8);
                    } else {
                        fileinfo->FirstCluster = (uint32_t)SD_L2_workBuf[j+0x1a] | (((uint32_t)SD_L2_workBuf[j+0x1b])<<8) 
                             | (((uint32_t)SD_L2_workBuf[j+0x14])<<16) | (((uint32_t)SD_L2_workBuf[j+0x15])<<24);
                    } 
                    
                    // Initialize some things
                    fileinfo->ActSector = SD_L2_Cluster2Sector(fileinfo->FirstCluster);
                    fileinfo->ActBytePos = 0;
                    return(0);
                }

            }
            
        }
    }
    if(SD_L2_FAT.PartType == SD_L2_PARTTYPE_FAT16) return(SD_L2_ERROR_FILE_NOT_FOUND);
    return(SD_L2_ERROR_DIR_EOC);
}

#if SD_ENABLE_DIR_VIEW
/**
 * Prints a directory.
 * Callback must point to a function that prints one string. 
 *
 * Works only over one cluster of directory information. If 
 * SD_L2_ERROR_DIR_EOC is returned call function again with next
 * cluster number. 
 *
 * Set cluster to 0 to access root directory.
 *
 * Deleted files and long name entries are not shown generally.
 *
 * Only files are printed that has their attributes set/unset regarding maskSet/maskUnset.
 * Examples for maskSet, maskUnset:
 *  Ouput everything:           0x00, 0x00
 *  Shortname files only:       0x00, 0x18
 *  Shortname files and dirs:   0x00, 0x08
 *  Shortname dirs:             0x10, 0x08
 *  Volume name:                0x08, 0x10
 *
 * Mask bits: B7 = 0, B6 = 0, B5 = archive, B4 = directory, 
 *            B3 = volume name, B2 = system, B1 = hidden, B0 = read only
 */
uint8_t SD_L2_Dir(const uint32_t cluster, const uint8_t maskSet, const uint8_t maskUnset, void (*callback)(char *entryLine))
{
    uint16_t maxsect = SD_L2_FAT.SecPerClus;
    uint32_t startsect = SD_L2_Cluster2Sector(cluster);
    char buf[16];
    
    if((SD_L2_FAT.PartType != SD_L2_PARTTYPE_FAT16) && (SD_L2_FAT.PartType != SD_L2_PARTTYPE_FAT32)) return(SD_L2_ERROR_FAT_NOT_INIT);
    if(cluster == 0) {
        // Set root dir sector
        startsect = SD_L2_FAT.RootDirStart;
        if(SD_L2_FAT.PartType == SD_L2_PARTTYPE_FAT16) maxsect = (uint16_t)((32 * (uint32_t)SD_L2_FAT.RootEntryCount + 511)/512);
    }
    
    for(uint16_t i = 0; i<maxsect; i++) {
        uint8_t retval = SD_L1_ReadBlock(startsect + i, SD_L2_workBuf);
        if(retval) return(retval);
        
        for(uint16_t j = 0; j<512; j+=32) {
            uint8_t attrib;
            if(SD_L2_workBuf[j] == 0) return(0);    // Last entry when first character of filename == 0
            if(SD_L2_workBuf[j] == 0xe5) continue;  // Skip deleted files
            if(SD_L2_workBuf[j] == 0x05) SD_L2_workBuf[j] = 0xE5;
            
            attrib = SD_L2_workBuf[j+0x0b];
            
            // Test masks (skip long file name entries also)
            if(((attrib & maskSet) == maskSet) && ((attrib & maskUnset) == 0) && (attrib != 0x0f)) {
                uint8_t z=0;
                // Prepare output
                for(uint16_t k = 0; k < 16; k++) buf[k] = ' ';
                for(uint16_t k = 0; k < 8; k++) {
                  buf[z] = SD_L2_workBuf[j+k];
                  if(buf[z] > ' ') z++;     // Remove space
                }
                buf[z++] = '.';
                for(uint16_t k = 0; k < 3; k++) buf[z++] = SD_L2_workBuf[j+k+0x08];
                //buf[z++] = ' ';
                buf[z++] = 0;

                callback(buf);
            }
            
        }
    }
    if(SD_L2_FAT.PartType == SD_L2_PARTTYPE_FAT16) return(0);
    return(SD_L2_ERROR_DIR_EOC);
}
#endif  /* SD_ENABLE_DIR_VIEW */
